from flask_restful.representations.json import output_json as _output_json
from flask import jsonify

def format_error(*errors):
    return {
        '_error': errors
    }

def output_json(data, code, headers=None):
    if type(data) is str:
        # Handle view returning a string
        message = data
        if code != 200:
            data = format_error(message)
        else:
            data = {'message': message}

    elif code == 405 and type(data) is dict and 'message' in data:
        # handle flask-restful default
        data = format_error(data.get('message'))

    return _output_json(data, code, headers)


def handle_bad_request(err):
    data = getattr(err, 'data')
    if data:
        # Get validations from the ValidationError object
        messages = data['exc'].messages
    else:
        messages = ['Invalid request']
    return jsonify(messages), 422




from raven import Client 
from raven.contrib.celery import register_signal, register_logger_signal
from raven.contrib.flask import Sentry


def create_client(conf, app_version='__UNKNOWN__'):
    """Creates a sentry client.
    @param app_version (string): App version sent to sentry for making events 
                                 more rich
    @param conf['SENTRY_DSN'] (string, required): DSN of sentry server
    @param conf['SENTRY_SITE'] (string): The site description of the deployment.
    """
    client = Client(
        conf['SENTRY_DSN'],
        site=conf.get('SENTRY_SITE'),
        release=app_version,
    )
    return client

def inject_sentry(app):
    """Injects sentry into a Flask Application
    
    Will only inject if SENTRY_DSN is specified.
    SENTRY_SITE and app.version is used to provide extra context to sentry events.

    @param app (Flask): A flask application to attach sentry to.
    """

    if app.config.get('SENTRY_DSN'):
        client = create_client(app.config, app.version)
        Sentry(app, client=client)


def celery_inject_sentry(celery):
    """Inject Sentry into a celery app. Requires raven.
    
    If SENTRY_DSN is specified in config, a sentry client is created and 
    attached. Additionally uses celery.version and SENTRY_SITE to provide extra
    context to sentry events.
    """
    if celery.conf.get('SENTRY_DSN'):
        client = create_client(celery.conf, 
                               app_version=getattr(celery, 'version', 'UNKNOWN'))

        register_logger_signal(client)
        register_signal(client)



from setuptools import setup
import os
import subprocess

PACKAGE_PATH = './twopi_flask_utils'

packages = []

for item in os.listdir(PACKAGE_PATH):
    path = os.path.join(PACKAGE_PATH, item)
    if not item.startswith('_') and os.path.isdir(path):
        packages.append(('twopi_flask_utils.' + item, path))

extras = {
    'ldap': [],
    'restful': ['flask-restful'],
    'celery': ['celery'],
    'sentry': ['raven[flask]']
}


tests_require = []
for x in extras.values():
    tests_require.extend(x)

def get_version():
    v = subprocess.check_output(['git', 'describe', '--dirty', '--always'])
    if type(v) == str:
        return v
    return v.decode('UTF-8')

setup(
    name='twopi-flask-utils',
    version=get_version().strip(),
    package_dir=dict(packages),
    packages=list(map(lambda x: x[0], packages)),
    extras_require=extras,
    tests_require=tests_require,
    test_suite='tests'
)



import json

class AppReqTestHelper(object):
    """
    Adds convenience self.post, self.get, self.put, self.delete, self.patch
    methods on the testcase object.

    Assumes a flask app client is defined on self.client
    """
    def _req(self, meth, *args, **kwargs):
        if kwargs.get('content_type') is None and meth != 'get':
            kwargs['content_type'] = 'application/json'

        if kwargs.get('json') is not None:
            kwargs['data'] = json.dumps(kwargs.get('json'))
            del kwargs['json']

        func = getattr(self.client, meth)
        rv = func(*args, **kwargs)

        def get_json():
            return json.loads(rv.data.decode('UTF-8'))

        rv.get_json = get_json
        return rv

    def post(self, *args, **kwargs):
        return self._req('post', *args, **kwargs)

    def get(self, *args, **kwargs):
        return self._req('get', *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._req('put', *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._req('delete', *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self._req('patch', *args, **kwargs)



class CRUDTestHelper(AppReqTestHelper):
    def do_crud_test(self, endpoint, data_1=None, data_2=None, key='id', 
                     check_keys=[], keys_from_prev=[], create=True, delete=True, 
                     update=True, read=True):

        if read:
            # Plural read on empty set.
            rv = self.get(endpoint)
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(len(rv.get_json()), 0)

        if create:
            # Create
            rv = self.post(endpoint, json=data_1)
            self.assertEqual(rv.status_code, 200)
            self.assertEqualDicts(check_keys, rv.get_json(), data_1)
            key_id = rv.get_json()[key]

        if read:
            # Plural read
            rv = self.get(endpoint)
            self.assertEqual(rv.status_code, 200)
            if create:
                self.assertEqual(len(rv.get_json()), 1)
                self.assertEqualDicts(check_keys, rv.get_json()[0], data_1)
            else:
                # No object should have been created
                self.assertEqual(len(rv.get_json()), 0)

        if read and create:
            # Singular Read
            rv = self.get(endpoint + '/' + str(key_id))
            self.assertEqual(rv.status_code, 200)
            prev_data = rv.get_json() # Keep this data so we can use it after update.
            self.assertEqualDicts(check_keys, prev_data, data_1)

        if update and create:
            # Singular Update
            rv = self.put(endpoint + '/' + str(key_id), json=data_2)
            self.assertEqual(rv.status_code, 200)
            self.assertEqualDicts(check_keys, rv.get_json(), data_2)
            self.assertEqualDicts(keys_from_prev, rv.get_json(), prev_data)
        
        if read and create:
            # Singular Read to confirm persisted.
            rv = self.get(endpoint + '/' + str(key_id))
            self.assertEqual(rv.status_code, 200)
            self.assertEqualDicts(check_keys, rv.get_json(), data_2)
            self.assertEqualDicts(keys_from_prev, rv.get_json(), prev_data)

        if delete and create:
            # Singular Deletion
            rv = self.delete(endpoint + '/' + str(key_id))
            self.assertEqual(rv.status_code, 200)

        if read and create:
            # Singular read on object which doesn't exist.
            rv = self.get(endpoint + '/' + str(key_id))
            self.assertEqual(rv.status_code, 404)

        if read:
            # Plural read on empty set
            rv = self.get(endpoint)
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(len(rv.get_json()), 0)

    def filteredDicts(self, keys, *dicts):
        ret = []
        for d in dicts:    
            d_filtered = dict((k, v) for k, v in d.items()
                              if k in keys)
            ret.append(d_filtered)
        
        return ret

    def assertEqualDicts(self, keys, d1, d2):
        self.assertEqual(*self.filteredDicts(keys, d1, d2))
